import * as pdfjsLib from 'pdfjs-dist'
import 'pdfjs-dist/build/pdf.worker.entry'
import { textTranslate } from './utils.js'
import { clearNode, createEl,} from '@common/utils/index.js'
import { escapeStringRegexp } from '@common/utils/index.js'
/*
  options
                    pdfjs 初始化参数
    url
    httpHeaders
    withCredentials
    cMapUrl
    cMapPacked
                  额外的参数
    container
    pageClassName
    canvasClassName
    textLayerClassName
    highlightClassName

*/


/*
  pdf 控制器 实现 pdf 的核心逻辑
*/
export class PDFController {
  static cache = {}
  static loadDocument(options) {
    const {
      url = '',
      httpHeaders = {},
      withCredentials = true,
      cMapUrl = import.meta.env.BASE_URL + '/cmaps/',
      cMapPacked = true
    } = options
    const cache = PDFController.cache
    const loadingTask = cache[url] || pdfjsLib.getDocument({
      url,
      withCredentials,
      httpHeaders,
      cMapUrl,
      cMapPacked
    })

    cache[url] = loadingTask
    return loadingTask
  }

  constructor(options) {
    this.$options = options
    this.ready = false
    this.pages = []
  }

  async init() {
    const loadingTask = PDFController.loadDocument(this.$options)
    const pdfDocument = await loadingTask.promise
    this.pdfDocument = pdfDocument
    clearNode(this.$options.container)
    this.ready = true
  }

  /*
    渲染 pdf 的原始页
  */
  async renderPage(page) {
    if (!this.ready) {
      throw new Error('pdfController 还未 ready')
    }
    const { container, pageClassName, canvasClassName } = this.$options

    const pageData = (this.pages[page] = this.pages[page] || { pageNum: page })

    if (!pageData.pageContainer) {
      const pdfPage = await this.pdfDocument.getPage(page)
      pageData.pdfPage = pdfPage

      pageData.pageContainer = createEl('div', {
        className: pageClassName,
        parent: container
      })
      pageData.canvas = createEl('canvas', {
        className: canvasClassName,
        parent: pageData.pageContainer
      })

      pageData.canvasContext = pageData.canvas.getContext('2d')
    }
    const { pdfPage, canvasContext, canvas, pageContainer } = pageData
    let scale = 1
    let viewport = pdfPage.getViewport({ scale })
    const pageWidth = pageContainer.offsetWidth
    scale = pageWidth / viewport.width
    if (scale !== 1) {
      viewport = pdfPage.getViewport({ scale: scale })
    }
    const deviceScale = window.devicePixelRatio || 1

    pageData.scale = scale
    pageData.viewport = viewport
    canvas.width = Math.floor(viewport.width * deviceScale)
    canvas.style.width = Math.floor(viewport.width) + 'px'
    canvas.height = Math.floor(viewport.height * deviceScale)
    canvas.style.height = Math.floor(viewport.height) + 'px'
    pdfPage.render({ canvasContext, viewport,
      transform: deviceScale !== 1 ? [deviceScale, 0, 0, deviceScale, 0, 0] : null
    })
  }

  /*
  获取 pdf page 的文本信息
*/
  async getPdfPageContent(pdfPage) {
    const content = await pdfPage.getTextContent()
    content.items = content.items.map(i => {
      return {
        ...i,
        str: textTranslate(i.str.replaceAll('...', '…'))
      }
    })
    return content
  }

  /*
    渲染 pdf 文本层
  */
  async renderTextLayer(page, layerName = 'default') {
    let pageData = this.pages[page]
    if (!pageData) {
      await this.renderPage(page)
      pageData = this.pages[page]
    }
    const { textLayerClassName } = this.$options
    const { pdfPage, pageContainer, viewport } = pageData
    let { textLayers } = pageData
    if (!textLayers) {
      textLayers = {}
      pageData.textLayers = textLayers
    }
    let layerData = textLayers[layerName]
    if (!layerData) {
      layerData = {}
      textLayers[layerName] = layerData
    }

    let textLayerContainer = layerData.textLayerContainer
    if (textLayerContainer) {
      clearNode(textLayerContainer)
    } else {
      textLayerContainer = createEl('div', {
        className: textLayerClassName,
        parent: pageContainer
      })
      layerData.textLayerContainer = textLayerContainer
    }
    const textContent = await this.getPdfPageContent(pdfPage)
    const textDivs = []

    pdfjsLib.renderTextLayer({
      textContent,
      viewport,
      container: textLayerContainer,
      textDivs
    })
    layerData.textContent = textContent
    layerData.textDivs = textDivs
  }

  /*
    文本层的高亮处理
    目前有两个实现方案
    1、纯前端匹配 + 渲染
      通过后端返回的文本信息进行匹配（进行了去空字符处理，后端 有可能加一些空字符 如 回车 空格等，可能影响匹配，所以统一去除空字符，在匹配，然后匹配到之后进行位置还原）这样可以跨句匹配 同时解决一些后端不稳定导致的匹配失败
    2、 后端测量好文本位置信息之后,返回 文本位置信息，前端根据位置信息绘制形状实现高亮效果

    目前使用的是 方案1  后续后端 ready 之后可以考虑使用 方案2
    方案1 需要前端考虑各种异常情况，如 pdf 由于字体问题导致的识别不准确（康熙字典、文字缺失、 … 识别为  . . . 等问题）
  */
  async highlightTextLayer(page, keyword, layerName = 'default') {
    if (!this.pages[page] || !this.pages[page].textLayers || !this.pages[page].textLayers[layerName]) {
      await this.renderTextLayer(page, layerName)
    }

    const layerData = this.pages[page].textLayers[layerName]
    let hightlightController = layerData.hightlightController
    if (!hightlightController) {
      layerData.highlightClassName = this.$options.highlightClassName
      hightlightController = new HighLightController(layerData)
      layerData.hightlightController = hightlightController
    }
    let matchSuccess = false
    matchSuccess = hightlightController.highLight(keyword)
    if (matchSuccess) return
    keyword = textTranslate(keyword)
    matchSuccess = hightlightController.highLight(keyword)
    if (matchSuccess) {
      console.log('keyword 中存在康熙部首 转换之后匹配正常')
      return
    }
    keyword = keyword.replaceAll('...', '…')
    matchSuccess = hightlightController.highLight(keyword)
    if (matchSuccess) {
      console.log('keyword 中存在 ... 转换为 … 之后匹配正常')
      return
    }
    console.log('匹配失败')
  }

}

/*
  纯前端的 文本匹配 高亮工具
*/
class HighLightController {
  static emptyCharReg = /\s+/ug
  static isEmpty = str => /^\s*$/u.test(str)
  static parseStr = str => {
    const empty = HighLightController.isEmpty(str)
    if (empty) {
      return {
        empty
      }
    }
    const emptyMatchs = []
    const parsedMatchs = []
    let tmpMatch
    // eslint-disable-next-line no-cond-assign
    while (tmpMatch = HighLightController.emptyCharReg.exec(str)) {
      emptyMatchs.push(tmpMatch)
    }
    const endIdx = str.length - 1
    let idx = 0
    let tmpStr
    let tmpLen
    while (idx <= endIdx) {
      tmpMatch = emptyMatchs.shift()
      if (!tmpMatch) {
        tmpStr = str.slice(idx)
        tmpLen = tmpStr.length
        idx = endIdx + 1
        parsedMatchs.push({
          str: tmpStr,
          len: tmpLen,
          empty: false
        })
        break
      }
      if (idx !== tmpMatch.index) {
        tmpStr = str.slice(idx, tmpMatch.index)
        tmpLen = tmpStr.length
        idx = tmpMatch.index
        parsedMatchs.push({
          str: tmpStr,
          len: tmpLen,
          empty: false
        })
      }
      tmpStr = tmpMatch[0]
      tmpLen = tmpStr.length
      idx += tmpLen
      parsedMatchs.push({
        str: tmpStr,
        len: tmpLen,
        empty: true
      })
    }
    const pureStr = parsedMatchs.filter(c => !c.empty).map(c => c.str).join('')
    return {
      parsedChars: parsedMatchs,
      pureStr,
      pureStrLen: pureStr.length,
      strLen: str.length
    }
  }
  creatSubNode({ text, isHighlight, parent }) {
    const { highlightClassName = 'highlight' } = this.$options
    let node
    if (isHighlight) {
      node = document.createElement('span')
      node.classList.add(highlightClassName)
      node.innerText = text
    } else {
      node = document.createTextNode(text)
    }

    if (parent) {
      parent.append(node)
    }
    return node
  }
  constructor(options) {
    this.$options = options
    this.init()
  }
  init() {
    const { textDivs, textContent } = this.$options
    const { items } = textContent
    if (items.length !== textDivs.length) {
      return `textContent.item 数量与 textDivs 数量不匹配
      请检查或者等待pdf textLayer 渲染完成之后在进行高亮解析
      textContent.item.length =  ${textContent.item.length}
      textDivs.length = ${textDivs.length}
      `
    }
    const newItems = []
    const pageContent = []
    for (let i = 0; i < items.length; i++) {
      const textItem = items[i]
      const textDiv = textDivs[i]
      const data = HighLightController.parseStr(textItem.str)
      if (data.pureStr) {
        pageContent.push(data.pureStr)
      }

      newItems.push({
        ...textItem,
        textDiv,
        ...data
      })
    }

    this.parsedData = {
      newItems,
      pageContent: pageContent.join('')
    }
  }

  resetTextLayer() {
    const { newItems } = this.parsedData
    for (let index = 0; index < newItems.length; index++) {
      const item = newItems[index]
      if (item.isHighlight) {
        clearNode(item.textDiv)
        this.creatSubNode({
          text: item.str,
          isHighlight: false,
          parent: item.textDiv
        })

        item.isHighlight = false
      }
    }
  }

  createRanges(keyword, pageContent) {
    if (!keyword || !pageContent) return
    const ranges = []
    const reg = new RegExp(escapeStringRegexp(keyword), 'gu')
    let startIdx = reg.lastIndex = 0
    let endIdx = 0
    let matchResult
    // debugger
    // eslint-disable-next-line no-cond-assign
    while (matchResult = reg.exec(pageContent)) {
      startIdx = endIdx
      endIdx = reg.lastIndex
      ranges.push(
        this.handleMatch({
          matchResult,
          startIdx,
          endIdx
        })
      )
    }

    return ranges
  }

  handleMatch({ matchResult, startIdx, endIdx }) {
    const { newItems } = this.parsedData
    const newRange = {
      startIdx, endIdx, matchResult
    }
    let idx = 0
    let itemIdx = 0
    let nextIdx = idx
    const startMatchIdx = matchResult.index
    while (idx <= startMatchIdx && itemIdx < newItems.length) {
      const item = newItems[itemIdx]
      if (!item.empty) {
        nextIdx = idx + item.pureStrLen
        if (nextIdx <= startMatchIdx) {
          idx += item.pureStrLen
        } else {
          newRange.startItem = item
          newRange.startOffset = startMatchIdx - idx
          newRange.startItemIdx = itemIdx
          newRange.startItemStartCharIdx = idx
          break
        }
      }
      itemIdx++
    }

    if (endIdx <= (newRange.startItemStartCharIdx + newRange.startItem.pureStrLen)) {
      newRange.endItem = newRange.startItem
      newRange.endItemIdx = newRange.startItemIdx
      newRange.endOffset = endIdx - newRange.startItemStartCharIdx
    }
    if (!newRange.endItem) {
      while (idx <= endIdx && itemIdx < newItems.length) {
        const item = newItems[itemIdx]
        if (!item.empty) {
          nextIdx = idx + item.pureStrLen
          if (nextIdx < endIdx) {
            idx += item.pureStrLen
          } else {
            newRange.endItem = item
            newRange.endItemIdx = itemIdx
            newRange.endOffset = endIdx - idx
            break
          }
        }
        itemIdx++
      }
    }
    return newRange
  }

  handleRanges(ranges) {
    let preRange
    let range
    let nextRange
    for (let i = 0; i < ranges.length; i++) {
      preRange = ranges[i - 1]
      range = ranges[i]
      nextRange = ranges[i + 1]
      this.handleSingRange({
        preRange,
        range,
        nextRange
      })
    }
  }

  handleSingRange({
    preRange,
    range,
    nextRange
  }) {
    const { newItems } = this.parsedData
    const {
      endItem, endOffset, endItemIdx,
      startItem, startOffset, startItemIdx
    } = range
    const startItemOffset = preRange && preRange.endItem === startItem ? preRange.endOffset : 0

    this.handleItem({
      item: startItem,
      startIdx: startItemOffset,
      endIdx: startOffset,
      isHighlight: false
    })

    if (startItemIdx === endItemIdx) {
      this.handleItem({
        item: startItem,
        startIdx: startOffset,
        endIdx: endOffset,
        isHighlight: true
      })
    } else {
      this.handleItem({
        item: startItem,
        startIdx: startOffset,
        endIdx: startItem.pureStrLen,
        isHighlight: true
      })

      let idx = startItemIdx + 1
      let item = newItems[idx]
      while (idx < endItemIdx) {
        item = newItems[idx]
        this.handleItem({
          item,
          startIdx: 0,
          endIdx: item.pureStrLen,
          isHighlight: true
        })
        idx++
      }

      this.handleItem({
        item: endItem,
        startIdx: 0,
        endIdx: endOffset,
        isHighlight: true
      })
    }
    if (endOffset !== endItem.pureStrLen) {
      const endItemOffset = nextRange && nextRange.startItem === endItem ? nextRange.startOffset : endItem.pureStrLen
      this.handleItem({
        item: endItem,
        startIdx: endOffset,
        endIdx: endItemOffset,
        isHighlight: false
      })
    }
  }
  handleItem({ item, startIdx, endIdx, isHighlight }) {
    const {
      parsedChars,
      empty
    } = item
    let idx = 0
    let nextIdx = 0
    let charIdx = 0
    let char
    let startCharIdx
    let endCharIdx
    let startOffset
    let endOffset
    let startChar
    if (empty) {
      return
    }
    if (!item.isHighlight) {
      clearNode(item.textDiv)
      let emptyIdx = 0
      let tmpChar
      while ((tmpChar = parsedChars[emptyIdx]).empty) {
        this.creatSubNode({
          text: tmpChar.str,
          isHighlight: false,
          parent: item.textDiv
        })
        emptyIdx++
      }
    }

    item.isHighlight = true
    while (idx <= startIdx && !startChar) {
      char = parsedChars[charIdx]
      if (char.empty) {
        charIdx++
        continue
      }
      nextIdx = idx + char.str.length
      if (nextIdx < startIdx) {
        idx = nextIdx
        charIdx++
      } else {
        startOffset = startIdx - idx
        startChar = char
        idx = startIdx
        startCharIdx = charIdx

        break
      }
    }

    if (nextIdx >= endIdx) {
      endCharIdx = charIdx
      endOffset = endIdx - idx + startOffset
    } else {
      // nextIdx = nextIdx - startOffset
      while (idx < endIdx) {
        char = parsedChars[charIdx]
        if (char.empty) {
          charIdx++
          continue
        }
        if (char === startChar) {
          nextIdx = idx + char.str.length - startOffset
        } else {
          nextIdx = idx + char.str.length
        }

        if (nextIdx < endIdx) {
          idx = nextIdx
          charIdx++
        } else {
          endOffset = endIdx - idx
          idx = endIdx
          endCharIdx = charIdx
          break
        }
      }
    }

    this.handleChars({
      parsedChars,
      endCharIdx,
      endOffset,
      startCharIdx,
      startOffset,
      item,
      isHighlight
    })
  }
  handleChars({
    parsedChars,
    endCharIdx,
    endOffset,
    startCharIdx,
    startOffset,
    item,
    isHighlight
  }) {
    const parent = item.textDiv
    let char = parsedChars[startCharIdx]
    if (startCharIdx === endCharIdx) {
      this.creatSubNode({
        text: char.str.slice(startOffset, endOffset),
        isHighlight,
        parent
      })
      return
    }
    this.creatSubNode({
      text: char.str.slice(startOffset, char.len),
      isHighlight,
      parent
    })

    let idx = startCharIdx + 1
    while (idx < endCharIdx) {
      char = parsedChars[idx]
      if (char.empty) {
        this.creatSubNode({
          text: char.str,
          isHighlight,
          parent
        })

        idx++
        continue
      }
      this.creatSubNode({
        text: char.str.slice(startOffset, char.len),
        isHighlight,
        parent
      })

      idx++
    }
    char = parsedChars[idx]
    this.creatSubNode({
      text: char.str.slice(0, endOffset),
      isHighlight,
      parent
    })
  }

  highLight(keyword) {
    const { pageContent } = this.parsedData
    const parsedKeywordData = HighLightController.parseStr(keyword)
    const pureKeyword = parsedKeywordData.pureStr
    if (!pageContent) return
    this.resetTextLayer()
    if (!pureKeyword) {
      return
    }

    const ranges = this.createRanges(pureKeyword, pageContent)

    /*
      匹配失败 将不匹配的文本 打印出来，进行排查问题
    */
    if (pageContent.indexOf(pureKeyword) === -1) {
      const chineseReg = /[\u4e00-\u9fa5|\u3002|\uff1f|\uff01|\uff0c|\u3001|\uff1b|\uff1a|\u201c|\u201d|\u2018|\u2019|\uff08|\uff09|\u300a|\u300b|\u3008|\u3009|\u3010|\u3011|\u300e|\u300f|\u300c|\u300d|\ufe43|\ufe44|\u3014|\u3015|\u2026|\u2014|\uff5e|\ufe4f|\uffe5]/
      for (let i = 0; i < pureKeyword.length; i++) {
        if (pureKeyword[i] === pageContent[i]) {
          continue
        }
        console.log(`idx  ${i}
                     i in keyword: ${pureKeyword[i]} isChines ${chineseReg.test(pureKeyword[i])}

                     i in pageContent ${pageContent[i]}  isChines ${chineseReg.test(pageContent[i])}
                     isEq ${String(pureKeyword[i] === pageContent[i])}
                     `)
      }
      console.log({
        pageContent,
        pureKeyword
      })
    }

    this.handleRanges(ranges)
    return ranges.length > 0
  }
}
